/*
* Licensed Materials - Property of IBM
* 5648-C63
* (C) Copyright IBM Corp. 1999, 2001
* This software includes contributions that are Copyright (c) 1998-2002 University of Chicago.
* All rights reserved.
*/

/******************************************************************************/
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/*   This source file implements a sample:                                    */
/*     - plugin node (Perl) which performs Perl computation                   */
/*     - framework for a plugin input node for illustrative purposes only     */
/*                                                                            */
/*                                                                            */
/* Implementation Notes:                                                      */
/*                                                                            */
/*   Portability:                                                             */
/*                                                                            */
/*     This code is intended to be portable and run on any of the operating   */
/*     systems supported by WebSphere MQ Integrator.                          */
/*                                                                            */
/******************************************************************************/

/* Includes */
/* <malloc.h> is not available on /390 */
#ifndef __MVS__
#include <malloc.h>
#endif
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <conio.h>
#include "PerlNode.h"


#define CONST_PERL_NODE                 "PerlNode"
const CciChar * constPerlNode          = 0;

#define CONST_TRACE_LOCATION       "PerlNode.trc"
const CciChar * constTraceLocation = 0;

#define CONST_PLUGIN_NODE_FACTORY            "PerlNodeFactory"
const CciChar * constPluginNodeFactory      = 0;

#define CONST_OUT                            "out"
const CciChar * constOut                    = 0;

#define CONST_ZERO                           "0"
const CciChar * constZero                   = 0;

#define CONST_NODE_TRACE_OUT_FILE            "nodeTraceOutfile"
const CciChar * constNodeTraceOutfile       = 0;






/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Display the environment in which a node is running                        */
/*                                                                            */
/******************************************************************************/
void displayEnvironment(
                        NODE_CONTEXT_ST* context
                        ){
   static char* functionName = (char *)"displayEnvironment()";
   CNI_BROKER_INFO_ST broker_info_st = {CNI_BROKER_INFO_ST_DEFAULT};
   
   /* Setup the BROKER_INFO_ST to query our environment */
   broker_info_st.brokerName.buffer               = (CciChar *)malloc(MAX_STRING_SIZE);
   broker_info_st.brokerName.bufferLength         = MAX_STRING_SIZE;
   broker_info_st.executionGroupName.buffer       = (CciChar *)malloc(MAX_STRING_SIZE);
   broker_info_st.executionGroupName.bufferLength = MAX_STRING_SIZE;
   broker_info_st.messageFlowName.buffer          = (CciChar *)malloc(MAX_STRING_SIZE);
   broker_info_st.messageFlowName.bufferLength    = MAX_STRING_SIZE;
   broker_info_st.queueManagerName.buffer         = (CciChar *)malloc(MAX_STRING_SIZE);
   broker_info_st.queueManagerName.bufferLength   = MAX_STRING_SIZE;
   
   /* Get details about the broker environment */
   cniGetBrokerInfo(0, ((NODE_CONTEXT_ST *)context)->nodeObject, &broker_info_st);
   
   /* Release storage */
   if (broker_info_st.brokerName.buffer)         free(broker_info_st.brokerName.buffer);
   if (broker_info_st.executionGroupName.buffer) free(broker_info_st.executionGroupName.buffer);
   if (broker_info_st.messageFlowName.buffer)    free(broker_info_st.messageFlowName.buffer);
   if (broker_info_st.queueManagerName.buffer)   free(broker_info_st.queueManagerName.buffer);
   
   return;
}



/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Acquire storage for attribute and link to the context                     */
/*                                                                            */
/******************************************************************************/
ATTRIBUTE_TBL_ENTRY *insAttrTblEntry(
                                     NODE_CONTEXT_ST* context,
                                     CciChar*         attrName,
                                     int              type
                                     ){
   static char* functionName = (char *)"insAttrTblEntry()";
   ATTRIBUTE_TBL_ENTRY* entry;
   int                  attrSize;
  

  
   entry = (ATTRIBUTE_TBL_ENTRY *)malloc(sizeof(ATTRIBUTE_TBL_ENTRY));
   if (entry) {
      entry->next = 0;
      entry->type = type;
      
      switch (entry->type) {
      case CNI_TYPE_INTEGER:
         attrSize = 4;
         break;
      case CNI_TYPE_STRING:
         attrSize = MAX_STRING_SIZE;

         break;
      default:
         attrSize = MAX_STRING_SIZE;
         /* Error: Invalid datatype */
         break;
      }
      
      /* Acquire and clear storage for attribute value */
      entry->valptr = malloc(attrSize);
      if (entry->valptr) memset(entry->valptr, 0, attrSize);
      entry->size=attrSize;
      /* Store the attribute name */
      CciCharCpy(entry->name, attrName);
      
      /* Link an existing previous element to this one */
      if (context->attrListPrevious) context->attrListPrevious->next = entry;
      else if ((context->attrListHead) == 0) context->attrListHead = entry;
      
      /* Save the pointer to the previous element */
      context->attrListPrevious = entry;
   }
   else {
      /* Error: Unable to allocate memory */
   }

           
   return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Return the terminal object for a named input terminal                     */
/*                                                                            */
/******************************************************************************/
void *getInputTerminalHandle(
                             NODE_CONTEXT_ST* context,
                             CciChar*         terminalName
                             ){
   static char* functionName = (char *)"getInputTerminalHandle()";
   TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->inputTerminalListHead;
   
   /* Scan the list for a matching terminal name */
   while (p != 0) {
      if (CciCharCmp(p->name, terminalName) == 0) {
         /* Match found; return it */
         return(p->handle);
      }
      p = (TERMINAL_LIST_ENTRY *)p->next;
   }
   
   /* No match was found */
   return(0);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Acquire storage for an input terminal entry and link it to the context    */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insInputTerminalListEntry(
                                               NODE_CONTEXT_ST* context,
                                               CciChar*         terminalName
                                               ){
   static char* functionName = (char *)"insInputTerminalListEntry()";
   TERMINAL_LIST_ENTRY* entry;
   int                  rc;
         
  
   entry = (TERMINAL_LIST_ENTRY *)malloc(sizeof(TERMINAL_LIST_ENTRY));
   if (entry) {
      
      /* This entry is the current end of the list */
      entry->next = 0;
      
      /* Store the terminal name */
      CciCharCpy(entry->name, terminalName);
      
      /* Create terminal and save its handle */
      entry->handle = cniCreateInputTerminal(&rc, context->nodeObject, (CciChar*)terminalName);
      
      /* Link an existing previous element to this one */
      if (context->inputTerminalListPrevious) context->inputTerminalListPrevious->next = entry;
      else if ((context->inputTerminalListHead) == 0) context->inputTerminalListHead = entry;
      
      /* Save the pointer to the previous element */
      context->inputTerminalListPrevious = entry;
   }
   else {
      /* Error: Unable to allocate memory */
   }

  
   return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Return the terminal object for a named output terminal                    */
/*                                                                            */
/******************************************************************************/
void *getOutputTerminalHandle(
                              NODE_CONTEXT_ST* context,
                              CciChar*         terminalName
                              ){
   static char* functionName = (char *)"getOutputTerminalHandle()";
   TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->outputTerminalListHead;
   
   /* Scan the list for a matching terminal name */
   while (p != 0) {
      if (CciCharCmp(p->name, terminalName) == 0) {
         
         /* Match found; return it */
         return(p->handle);
      }
      p = (TERMINAL_LIST_ENTRY *)p->next;
   }
   
   /* No match was found */
   return(0);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Acquire storage for an output terminal and link it to the context         */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insOutputTerminalListEntry(
                                                NODE_CONTEXT_ST* context,
                                                CciChar*         terminalName
                                                ){
   static char* functionName = (char *)"insOutputTerminalListEntry()";
   TERMINAL_LIST_ENTRY* entry;
   int                  rc;

     
   entry = (TERMINAL_LIST_ENTRY *)malloc(sizeof(TERMINAL_LIST_ENTRY));
   if (entry) {
      
      /* This entry is the current end of the list */
      entry->next = 0;
      
      /* Store the terminal name */
      CciCharCpy(entry->name, terminalName);
      
      /* Create terminal and save its handle */
      entry->handle = cniCreateOutputTerminal(&rc, context->nodeObject, (CciChar*)terminalName);
      
      /* Link an existing previous element to this one */
      if (context->outputTerminalListPrevious) context->outputTerminalListPrevious->next = entry;
      else if ((context->outputTerminalListHead) == 0) context->outputTerminalListHead = entry;
      
      /* Save the pointer to the previous element */
      context->outputTerminalListPrevious = entry;
   }
   else {
      /* Error: Unable to allocate memory */
   }

  
   return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniCreateNodeContext() */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniCreateNodeContext() */
/*                                                                            */
/******************************************************************************/
CciContext* _Perl_createNodeContext(
                                    CciFactory* factoryObject,
                                    CciChar*    nodeName,
                                    CciNode*    nodeObject
                                    ){
   static char* functionName = (char *)"_Perl_createNodeContext()";
   NODE_CONTEXT_ST* p;
  
   /* Allocate a pointer to the local context */
   p = (NODE_CONTEXT_ST *)malloc(sizeof(NODE_CONTEXT_ST));
   
   if (p) {
      
      /* Clear the context area */
      memset(p, 0, sizeof(NODE_CONTEXT_ST));
      
      /* Save our node object pointer in our context */
      p->nodeObject = nodeObject;

      /* Save our node name */
      CciCharNCpy((CciChar*)&p->nodeName, nodeName, MAX_NODE_NAME_LEN);
      
      /* Create attributes and set default values */
         {
         const CciChar* ucsAttr = CciString("perlScript", BIP_DEF_COMP_CCSID) ;
         insAttrTblEntry(p, (CciChar*)ucsAttr , CNI_TYPE_STRING);
         free((void *)ucsAttr) ;
         }
         
         /* Create terminals */
         {
         const CciChar* ucsIn = CciString("in", BIP_DEF_COMP_CCSID) ;
         insInputTerminalListEntry(p, (CciChar*)ucsIn);
         free((void *)ucsIn) ;
         }
         
         {
         const CciChar* ucsOut = CciString("out", BIP_DEF_COMP_CCSID) ;
         insOutputTerminalListEntry(p, (CciChar*)ucsOut);
         free((void *)ucsOut) ;
         }
         
         {
         
        
         /* Initialise default trace setting */
         p->trace = 0; 
         
         /* If tracing, open the file */
         if (p->trace) {
            const char * filename = "C:\\perl.trc";
            p->tracefile = fopen(filename, "w+");
         }
         }
   }
   else {
      /* Error: Unable to allocate memory */
   }

  
   /* Return the pointer to the local context */
   return((CciContext*)p);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniDeleteNodeContext() */
/*                                                                            */
/* This function is called when an instance of a node is deleted.             */
/*                                                                            */
/******************************************************************************/
void _deleteNodeContext(
                        CciContext* context
                        ){
   static char* functionName = (char *)"_deleteNodeContext()";
   
   return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int _getAttributeName(
                      CciContext* context,
                      int         index,
                      CciChar*    buffer,
                      int         bufsize
                      ){
   static char* functionName = (char *)"_getAttributeName()";
   int                  count = 0;

   ATTRIBUTE_TBL_ENTRY *p = ((NODE_CONTEXT_ST *)context)->attrListHead;

  
   while (p != 0) {
      if (count == index) {
         CciCharCpy(buffer, p->name);
         
         /* Normal return; attribute name returned */
         return(0);
      }
      count++;
      p = (ATTRIBUTE_TBL_ENTRY *)p->next;
   }

  
   /* Attribute not owned by the plugin node; defer to base */
   return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniSetAttribute() */
/*                                                                            */
/******************************************************************************/
int _setAttribute(
                  CciContext* context,
                  CciChar*    attrName,
                  CciChar*    attrValue
                  ){
   static char* functionName = (char *)"_setAttribute()";
   ATTRIBUTE_TBL_ENTRY* p = ((NODE_CONTEXT_ST *)context)->attrListHead;
   NODE_CONTEXT_ST*     nc = (NODE_CONTEXT_ST *)context;

           
   if (nc->trace) {
      const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
      const char *mbAttrName = mbString(attrName, BIP_DEF_COMP_CCSID);
      const char *mbAttrValue = mbString(attrValue, BIP_DEF_COMP_CCSID);
      fprintf(nc->tracefile,
         (char *)"NODE: <%s> -> %s context=0x%x attrName='%s' attrValue='%s'\n",
         mbNodeName, functionName, context, mbAttrName, mbAttrValue);
      fflush(nc->tracefile);
      free((void *)mbNodeName);
      free((void *)mbAttrName);
      free((void *)mbAttrValue);
   }
   
   while (p != 0) {
      if (CciCharCmp(p->name, attrName) == 0) {
         
         switch(p->type) {
         case CNI_TYPE_INTEGER:
            {
            const char* mbAttrValue = mbString(attrValue, BIP_DEF_COMP_CCSID);
            *((int *)p->valptr) = atoi(mbAttrValue);
            free((void *)mbAttrValue);
            }
            
            break;
         case CNI_TYPE_STRING:
            {
            const char* mbAttrValue = mbString(attrValue, BIP_DEF_COMP_CCSID);
            memcpy(p->valptr, mbAttrValue, MAX_STRING_SIZE);
            free((void *)mbAttrValue);
            }
            break;
            
         default:
            /* Error: Invalid datatype */
            break;
         }
         
         if (nc->trace) {
            const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
            fprintf( nc->tracefile, (char *)"NODE: <%s> <- %s rc=0\n",
               mbNodeName, functionName);
            fflush(nc->tracefile);
            free((void *)mbNodeName);
         }
         
         /* Normal return; attribute value stored by plugin */

  
 
         return(0);
      }
      p = (ATTRIBUTE_TBL_ENTRY *)p->next;
   }
   
   if (nc->trace) {
      const char* mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
      fprintf(nc->tracefile, (char *)"NODE: <%s> <- %s rc=1\n",
         mbNodeName, functionName);
      fflush(nc->tracefile);
      free((void *)mbNodeName);
   }

  
   /* Attribute not owned by plugin */
   return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int _getAttribute(
                  CciContext* context,
                  CciChar*    attrName,
                  CciChar*    buffer,
                  int         bufsize
                  ){
   static char* functionName = (char *)"_getAttribute()";
   ATTRIBUTE_TBL_ENTRY* p = ((NODE_CONTEXT_ST *)context)->attrListHead;
   NODE_CONTEXT_ST*     nc = (NODE_CONTEXT_ST *)context;
   char                 string[256];

 
   if (nc->trace) {
      const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
      const char *mbAttrName = mbString(attrName, BIP_DEF_COMP_CCSID);
      fprintf(nc->tracefile, "NODE: <%s> -> %s context=0x%x attrName='%s'\n",
         mbNodeName, functionName, context, mbAttrName);
      fflush(nc->tracefile);
      free((void *)mbNodeName);
      free((void *)mbAttrName);
   }
   
   while (p != 0) {
      if (CciCharCmp(p->name, attrName) == 0) {
         
         switch(p->type) {
         case CNI_TYPE_INTEGER:
            sprintf((char*)&string, "%d", *((int*)p->valptr));
               {
               const CciChar* ucsAttrValue = CciString(string, BIP_DEF_COMP_CCSID); /*TODO*/
               CciCharNCpy(buffer, ucsAttrValue, min(bufsize, 256));
               free((void *)ucsAttrValue);
               }
               break;
         case CNI_TYPE_STRING:
            {
            const CciChar* ucsAttrValue = CciString((char *)p->valptr, BIP_DEF_COMP_CCSID); /* TODO */
            CciCharNCpy(buffer, ucsAttrValue, min(p->size, bufsize));
            free((void *)ucsAttrValue);
            }
         default:
            /* Error: Invalid datatype */
            break;
         }
         
         if (nc->trace) {
            const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
            fprintf(nc->tracefile, "NODE: <%s> <- %s rc=0\n",
               mbNodeName, functionName);
            fflush(nc->tracefile);
            free((void *)mbNodeName);
         }
         
         /* Normal return; attribute value returned */
         return(0);
      }
      p = (ATTRIBUTE_TBL_ENTRY *)p->next;
   }
   
   if (nc->trace) {
      const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
      fprintf(nc->tracefile, "NODE: <%s> <- %s rc=1\n",
         mbNodeName, functionName);
      fflush(nc->tracefile);
      free((void *)mbNodeName);
   }

  
   /* Attribute not owned by plugin */
   return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
char * _getPerlScript(
                      CciContext* context
                      ){
   static char* functionName = (char *)"_getPerlScript()";
   ATTRIBUTE_TBL_ENTRY* p = ((NODE_CONTEXT_ST *)context)->attrListHead;
   NODE_CONTEXT_ST*     nc = (NODE_CONTEXT_ST *)context;
   char           *     perlScript=0;

  
   if (nc->trace) {
      const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
      fprintf(nc->tracefile, "GetPerlScript NODE: <%s> -> %s context=0x%x \n",mbNodeName, functionName, context);
      fflush(nc->tracefile);
      free((void *)mbNodeName);
   }
   
   while (p != 0) {
      if (CciCharCmp(p->name, L"perlScript") == 0) {
         if (p->type==CNI_TYPE_STRING)   {
            perlScript=(char *)p->valptr;
            break;
         }
         if (nc->trace) {
            const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
            fprintf(nc->tracefile, "NODE: <%s> <- %s rc=0\n",
               mbNodeName, functionName);
            fflush(nc->tracefile);
            free((void *)mbNodeName);
         }
      }
      p = (ATTRIBUTE_TBL_ENTRY *)p->next;
   }
   
   if (nc->trace) {
      const char *mbNodeName = mbString(nc->nodeName, BIP_DEF_COMP_CCSID);
      fprintf(nc->tracefile, "NODE: <%s> <- %s rc=1\n",mbNodeName, functionName);
      fflush(nc->tracefile);
      free((void *)mbNodeName);
   }

  
   /* Attribute not owned by plugin */
   return(perlScript);
}
/******************************************************************************/
/*                                                                            */
/* Perl  dynamc loading of wmqi extensions                                    */
/*                                                                            */
/******************************************************************************/
EXTERN_C void boot_DynaLoader (pTHX_ CV* cv);
EXTERN_C void  boot_wmqi (pTHX_ CV* cv);

EXTERN_C void
xs_init(pTHX)
   {
   char *file = __FILE__;
   /* DynaLoader is a special case */
   newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
   newXS("wmqi::bootstrap", boot_wmqi, file);
}
/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                         cniEvaluate() */
/*                                                                            */
/******************************************************************************/
void _Perl_evaluate(
                    CciContext* nodeContext,
                    CciMessage* destinationList,
                    CciMessage* exceptionList,
                    CciMessage* inputMsg
                    ){
   WmqiContext ctx;
   char *embedding[] = { "", "-e", "0" };
   char *parsecheck[] = { "", "-e", "0" };
   char *perlScript=0;
   char *perlScriptName="";
   int  rc;
   PerlInterpreter *my_perl;
   char * buf = "";
      
   perlScriptName=_getPerlScript(nodeContext);

   ctx.nodeContext=nodeContext;
   ctx.destinationList=destinationList;
   ctx.exceptionList=exceptionList;
   ctx.inputMsg=inputMsg;
   ctx.outputTerminalObject = getOutputTerminalHandle(nodeContext, L"out"); 

   
   
   perlScript=readPerlScript(perlScriptName);
   
   if ((ctx.outputTerminalObject!=0) && (perlScript!=0) ) {
      SV * sv; // Perl reference 
      
      // Allocate perl interpreter
      my_perl = perl_alloc();
      parsecheck[0]="";
      parsecheck[1]=perlScriptName;
      // Construct perl interpreter
      PERL_SET_CONTEXT(my_perl);
      perl_construct( my_perl );
      PERL_SET_CONTEXT(my_perl);
      rc= perl_parse(my_perl,xs_init, 2, parsecheck, NULL);
      printf("parse rc %d\n",rc);
      // Put the node context in a perl reference named nodeContext
      PERL_SET_CONTEXT(my_perl);	
      sv=get_sv("wmqi::inputMessage", TRUE);
      
      PERL_SET_CONTEXT(my_perl);
      sv_setref_pv(sv, "_p_CciMessage",inputMsg);

      PERL_SET_CONTEXT(my_perl);
      sv=get_sv("wmqi::context", TRUE);
      PERL_SET_CONTEXT(my_perl);
      sv_setref_pv(sv, "_p_WmqiContext",&ctx);

      /** Evaluate perl script attribute **/
      PERL_SET_CONTEXT(my_perl);
      perl_eval_pv("use wmqi;", FALSE);

      PERL_SET_CONTEXT(my_perl);
      perl_run(my_perl);
      
      

      PERL_SET_CONTEXT(my_perl);
      perl_destruct(my_perl);

      PERL_SET_CONTEXT(my_perl);
      perl_free(my_perl);
      
   } else {
      /* Should raise exception here */
   }
   if (perlScript) free((void *)perlScript);

   printf("<_Perl_evaluate\n"); 
           
   return;
}




/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Node definition - PerlNode                                                */
/*                                                                            */
/******************************************************************************/
void definePerlNode(
                    void* factoryObject
                    ){
   
   static char* functionName = (char *)"definePerlNode()";
   static CNI_VFT vftable = {CNI_VFT_DEFAULT};
   int            rc = 0;
 
   
   /* Setup function table with pointers to node implementation functions */
   vftable.iFpCreateNodeContext = _Perl_createNodeContext;
   vftable.iFpDeleteNodeContext = _deleteNodeContext;
   vftable.iFpGetAttributeName  = _getAttributeName;
   vftable.iFpSetAttribute      = _setAttribute;
   vftable.iFpGetAttribute      = _getAttribute;
   vftable.iFpEvaluate          = _Perl_evaluate;
   
   /* Define a node type supported by our factory. If any errors/exceptions    */
   /* occur during the execution of this utility function, then as we have not */
   /* supplied the returnCode argument, the exception will bypass the plugin   */
   /* and be directly handled by the broker.                                   */
   cniDefineNodeClass(&rc, factoryObject, (CciChar*)constPerlNode, &vftable);
   
   return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  initNodeConstants                                                         */
/*                                                                            */
/*  Creates Cci constant strings from defined char * constant strings.        */
/*                                                                            */
/******************************************************************************/
void initNodeConstants(
                       void
                       ){
   constZero                   = CciString(CONST_ZERO, BIP_DEF_COMP_CCSID);
   constTraceLocation          = CciString(CONST_TRACE_LOCATION, BIP_DEF_COMP_CCSID);
   constPerlNode               = CciString(CONST_PERL_NODE, BIP_DEF_COMP_CCSID);
   constPluginNodeFactory      = CciString(CONST_PLUGIN_NODE_FACTORY, BIP_DEF_COMP_CCSID);
   constNodeTraceOutfile       = CciString(CONST_NODE_TRACE_OUT_FILE, BIP_DEF_COMP_CCSID);
   constOut                    = CciString(CONST_OUT, BIP_DEF_COMP_CCSID);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                                       */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine.                                       */
/*                                                                            */
/* The responsibilities of the plug-in are to:                                */
/*  - create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/*  - define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/*  - define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/*  - define input and output terminals supported by each node.               */
/*  - return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */
/*                                                                            */
/******************************************************************************/
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
   {
   static char* functionName = (char *)"bipGetMessageflowNodeFactory()";
   CciFactory*      factoryObject;
   int              rc = 0;
   CCI_EXCEPTION_ST exception_st = {CCI_EXCEPTION_ST_DEFAULT};
   
   /* Before we proceed we need to initialise all the static constants */
   /* that may be used by the plug-in.                                 */

   initNodeConstants();
   
   /* Create the Node Factory for this plug-in */
   factoryObject = cniCreateNodeFactory(0, (unsigned short *)constPluginNodeFactory);
   if (factoryObject == CCI_NULL_ADDR) {
      if (rc == CCI_EXCEPTION) {
         /* Get details of the exception */
         memset(&exception_st, 0, sizeof(exception_st));
         cciGetLastExceptionData(&rc, &exception_st);
         
         /* Any local error handling may go here */
         
         /* Rethrow the exception */
         cciRethrowLastException(&rc);
      }
      /* Any further local error handling can go here */
   }
   else {
      /* Define the nodes supported by this factory */

      definePerlNode(factoryObject);

   }
   
   /* Return address of this factory object to the broker */
   return(factoryObject);
}

/*********************************************************************\
*  function:  FindElement()
*
*  Description:    
*
\*********************************************************************/
CciElement*   FindElement(char * varname,
                          NODE_CONTEXT_ST* context,
                          CciMessage* message,
                          int create
                          )
   {
   CciElement*      InRootElement;
   CciElement*      Element;
   CciElement*      NextElement;
   int               rc;
   int               found,strLen;
   char *            ptr1;
   char *            ptr2;
   int               elementNameSize;
   char              NextElementName[ELT_NAME_LEN];
   wchar_t           *wNextElementName;
   wchar_t           elementName[ELT_NAME_LEN];
   /* Get the root element of the message */
   InRootElement = cniRootElement(&rc, message);
   Element = InRootElement;
   found = 0;
   
   ptr1 = strstr(varname, ".");
   while (ptr1) {
      ptr1++; /* skip dot */
      ptr2 = strstr(ptr1, "."); /* find next dot */
      if (ptr2)
         strLen = (int) (ptr2-ptr1);
      else
         strLen = strlen(varname) - (int) (ptr1 - varname);
      memcpy(NextElementName, ptr1, strLen);
      NextElementName[strLen] = '\0';
      _strlwr(NextElementName);
      wNextElementName = ASCIIToUnicode(NextElementName);
      /* Go to the first child of this element */
      NextElement = cniFirstChild(&rc, Element);
      
      /* Loop until element found or no elements left */
      found = 0;
      
      while (NextElement) {
         /* Get the element's name */
         elementNameSize = cniElementName(&rc, NextElement, (CciChar*)&elementName, sizeof(elementName));
         /* Check if name is right */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         if (CciCharCmp(elementName, wNextElementName) == 0) 					/* 22/07/2001 */
#else																									/* 22/07/2001 */
            if ( wcscmp(elementName, wNextElementName) == 0 ) 
#endif																								/* 22/07/2001 */
               {   
               found = 1;
               break;
            }
            NextElement = cniNextSibling(&rc, NextElement);
      }
      
      if (found == 1) {
         /* Element found: go to this child */
         Element = NextElement;
         NextElement=0; /* break from while NextElement loop go one level deeper */
      }
      else {
         /* Element not found: break from while ptr1 loop */
         found = 0;
         break;
      }
      
      ptr1 = ptr2;
   }
   /* varname path does not exists or is incomplete create the additional elements */   
   if (create == 1) {
      
      
   }
   return 0;
}
