
#include <memory.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include "perlnode.h"

void setValueType(CciElement *targetElement) {
   CciElementType eltType;   
   int  rc;
   eltType = cniElementType(&rc, targetElement);   
   eltType&=CCI_ELEMENT_TYPE_MASK_GENERIC;
   eltType|=CCI_ELEMENT_TYPE_VALUE;
   cniSetElementType(&rc, targetElement, eltType);
}


/*Get the root element for a message */
CciElement  *RootElement(
                         WmqiContext *ctx,
                         CciMessage   *message
                         ){
   
   return cniRootElement(&ctx->retcode,message);
}


/*Create an element before the specified element */
CciElement  *CreateElementBefore(
                                 WmqiContext *ctx,
                                 CciElement *targetElement
                                 ){
   return cniCreateElementBefore(&ctx->retcode,targetElement);
}

/*Create an element before the specified element */
CciElement  *CreateElementBeforeUsingParser(
                                            WmqiContext *ctx,
                                            CciElement *targetElement,
                                            char   *parserClassName
                                            ){
   CciElement *elt;
   const CciChar * wparserClassName;
   wparserClassName=CciString(parserClassName, BIP_DEF_COMP_CCSID) ;
   elt=cniCreateElementBeforeUsingParser(&ctx->retcode,targetElement,wparserClassName);
   free((void *)wparserClassName);
   return elt;
}

/*Create an element after the specified element */
CciElement  *CreateElementAfter(
                                WmqiContext *ctx,
                                CciElement *targetElement
                                ){
   return cniCreateElementAfter(&ctx->retcode,targetElement);
}

/*Create an element after the specified element */
CciElement  *CreateElementAfterUsingParser(
                                           WmqiContext *ctx,
                                           CciElement *targetElement,
                                           char   *parserClassName
                                           ){
   CciElement *elt;
   const CciChar * wparserClassName;
   wparserClassName=CciString(parserClassName, BIP_DEF_COMP_CCSID) ;
   elt=    cniCreateElementAfterUsingParser(&ctx->retcode,targetElement,wparserClassName);
   free((void *)wparserClassName);
   return elt;
   
}

/*Create an element as a first child of the specified element */
CciElement  *CreateElementAsFirstChild(
                                       WmqiContext *ctx,
                                       CciElement *targetElement
                                       ){
   return cniCreateElementAsFirstChild(&ctx->retcode,targetElement);
}

/*Create an element as a first child of the specified element */
CciElement  *CreateElementAsFirstChildUsingParser(
                                                  WmqiContext *ctx,
                                                  CciElement *targetElement,
                                                  char   *parserClassName
                                                  ){
   CciElement *elt;
   const CciChar * wparserClassName;
   wparserClassName=CciString(parserClassName, BIP_DEF_COMP_CCSID) ;
   elt=   cniCreateElementAsFirstChildUsingParser(&ctx->retcode,targetElement,wparserClassName);
   free((void *)wparserClassName);
   return elt;
   
}

/*Create an element as a last child of the specified element */
CciElement  *CreateElementAsLastChild(
                                      WmqiContext *ctx,
                                      CciElement *targetElement
                                      ){
   return cniCreateElementAsLastChild(&ctx->retcode,targetElement);
}

/*Create an element as a last child of the specified element */
CciElement  *CreateElementAsLastChildUsingParser(
                                                 WmqiContext *ctx,
                                                 CciElement *targetElement,
                                                 char   *parserClassName
                                                 ){
   
   CciElement *elt;
   const CciChar * wparserClassName;
   wparserClassName=CciString(parserClassName, BIP_DEF_COMP_CCSID) ;
   elt=   cniCreateElementAsLastChildUsingParser(&ctx->retcode,targetElement,wparserClassName);
   free((void *)wparserClassName);
   return elt;
}

/*Add an element before a target element */
void   AddBefore(
                 WmqiContext *ctx,
                 CciElement *targetElement,
                 CciElement *newElement
                 ){
   cniAddBefore(&ctx->retcode,targetElement,newElement);
}

/*Add an element after a target element */
void   AddAfter(
                WmqiContext *ctx,
                CciElement *targetElement,
                CciElement *newElement
                ){
   cniAddAfter(&ctx->retcode,targetElement,newElement);
}

/*Add an element as a first child of a target element */
void   AddAsFirstChild(
                       WmqiContext *ctx,
                       CciElement *targetElement,
                       CciElement *newElement
                       ){
   cniAddAsFirstChild(&ctx->retcode,targetElement,newElement);
}

/*Add an element as a last child of a target element */
void   AddAsLastChild(
                      WmqiContext *ctx,
                      CciElement *targetElement,
                      CciElement *newElement
                      ){
   cniAddAsLastChild(&ctx->retcode,targetElement,newElement);
}

/*Detach the target element */
void   Detach(
              WmqiContext *ctx,
              CciElement *targetElement
              ){
   cniDetach(&ctx->retcode,targetElement);
}

/*Return the parent of a target element */
CciElement  *Parent(
                    WmqiContext *ctx,
                    CciElement *targetElement
                    ){
   return cniParent(&ctx->retcode,targetElement);
}

/*Return the previous sibling of a target element */
CciElement  *PreviousSibling(
                             WmqiContext *ctx,
                             CciElement *targetElement
                             ){
   return cniPreviousSibling(&ctx->retcode,targetElement);
}

/*Return the next sibling of a target element */
CciElement  *NextSibling(
                         WmqiContext *ctx,
                         CciElement *targetElement
                         ){
   CciElement *next=0;
   next=cniNextSibling(&ctx->retcode,targetElement);
   return next;
}

/*Return the first child of a target element */
CciElement  *FirstChild(
                        WmqiContext *ctx,
                        CciElement *targetElement
                        ){
   return cniFirstChild(&ctx->retcode,targetElement);
}

/*Return the last child of a target element */
CciElement  *LastChild(
                       WmqiContext *ctx,
                       CciElement *targetElement
                       ){
   return cniLastChild(&ctx->retcode,targetElement);
}

/*Set the type attribute of a target element */
void   SetElementType(
                      WmqiContext *ctx,
                      CciElement *targetElement,
                      int type
                      ){
   cniSetElementType(&ctx->retcode,targetElement,type);
}

/*Set the name attribute of a target element */
void   SetElementName(
                      WmqiContext *ctx,
                      CciElement *targetElement,
                      char   *name
                      ){
   const CciChar * wname;
   int  rc;
   CciElementType eltType;   
   wname=CciString(name, BIP_DEF_COMP_CCSID) ;
   eltType = cniElementType(&rc, targetElement);   
   eltType|=CCI_ELEMENT_TYPE_NAME;
   
   cniSetElementName(&ctx->retcode,targetElement,wname);
   free((void *)wname);
}

/*Set the value attribute of a target element */
void   SetElementValueValue(
                            WmqiContext *ctx,
                            CciElement *targetElement,
                            CciElementValue*value
                            ){
   cniSetElementValueValue(&ctx->retcode,targetElement,value);
}

/*Set the value of a target element */
void   SetElementBooleanValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              int     value
                              ){
   setValueType(targetElement);
   cniSetElementBooleanValue(&ctx->retcode,targetElement,(char)value);
}

/*Set the value of a target element */
void   SetElementIntegerValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              CciInt      value
                              ){
   setValueType(targetElement);
   cniSetElementIntegerValue(&ctx->retcode,targetElement,value);
}

/*Set the value of a target element */
void   SetElementRealValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           double     value){
   setValueType(targetElement);
   cniSetElementRealValue(&ctx->retcode,targetElement,value);
}

/*Set the value of a target element */
void   SetElementCharacterValue(
                                WmqiContext *ctx,
                                CciElement *targetElement,
                                char   *value,
                                int        length
                                ){
   const CciChar * wvalue;
   setValueType(targetElement);
   wvalue=CciString(value, BIP_DEF_COMP_CCSID) ;
   cniSetElementCharacterValue(&ctx->retcode,targetElement,wvalue,length); 
   free((void *)wvalue);
}

/*Set the value of a target element */
void   SetElementDateValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           int year, int month, int day
                           ){
   struct CciDate value;
   value.day=day;
   value.month=month;
   value.year=year;
   setValueType(targetElement);
   cniSetElementDateValue(&ctx->retcode,targetElement,&value);
}

/*Set the value of a target element */
void   SetElementTimeValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           int hour, int minute, float second
                           ){
   struct CciTime value;
   value.hour=hour;
   value.minute=minute;
   value.second=second;
   setValueType(targetElement);
   cniSetElementTimeValue(&ctx->retcode,targetElement,&value);
}

/*Set the value of a target element */
void   SetElementByteArrayValue(
                                WmqiContext *ctx,
                                CciElement *targetElement,
                                const struct CciByteArray*value
                                ){
   setValueType(targetElement);
   cniSetElementByteArrayValue(&ctx->retcode,targetElement,value);
}

/*Set the value of a target element */
void   SetElementTimestampValue(
                                WmqiContext *ctx,
                                CciElement *targetElement,
                                int year, int month, int day, int hour, int minute, float second
                                )
   {
   struct CciTimestamp value;
   value.day=day;
   value.month=month;
   value.year=year;
   value.hour=hour;
   value.minute=minute;
   value.second=second;
   setValueType(targetElement);
   cniSetElementTimestampValue(&ctx->retcode,targetElement,&value);
}

/*Set the value of a target element */
void   SetElementGmtTimestampValue(
                                   WmqiContext *ctx,
                                   CciElement *targetElement,
                                   int year, int month, int day, int hour, int minute, float second
                                   ){
   struct CciTimestamp value;
   value.day=day;
   value.month=month;
   value.year=year;
   value.hour=hour;
   value.minute=minute;
   value.second=second;
   setValueType(targetElement);
   cniSetElementGmtTimestampValue(&ctx->retcode,targetElement,&value);
}

/*Set the value of a target element */
void   SetElementGmtTimeValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              int hour, int minute, float second
                              ){
   struct CciTime value;
   value.hour=hour;
   value.minute=minute;
   value.second=second;
   setValueType(targetElement);
   cniSetElementGmtTimeValue(&ctx->retcode,targetElement,&value);
}

/*Set the value of a target element */
void   SetElementDecimalValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              char   *value
                              )   {
   const CciChar * wvalue;
   wvalue=CciString(value, BIP_DEF_COMP_CCSID) ;
   setValueType(targetElement);
   cniSetElementDecimalValue(&ctx->retcode,targetElement,wvalue);
   free((void *)wvalue);
}

/*Set the value of a target element */
void   SetElementBitArrayValue(
                               WmqiContext *ctx,
                               CciElement *targetElement,
                               const struct CciBitArray*value
                               ){
   setValueType(targetElement);
   cniSetElementBitArrayValue(&ctx->retcode,targetElement,value);
}

/*Return the type attribute of a target element */
int   ElementType(
                  WmqiContext *ctx,
                  CciElement *targetElement
                  ){
   return cniElementType(&ctx->retcode,targetElement);
}

/*Return the name attribute of a target element */
char *   ElementName(
                     WmqiContext *ctx,
                     CciElement *targetElement
                     ){
   int length=INIT_CHAR_LENGTH;
   const CciChar * wvalue;
   char *value;
   wvalue=(CciChar * )malloc((length+1)*sizeof(CciChar));
   memset((void *)wvalue,0x00,(length+1)*sizeof(CciChar));
   length= cniElementName(&ctx->retcode,targetElement,wvalue,length);
   if ((length>INIT_CHAR_LENGTH)&&(ctx->retcode==CCI_BUFFER_TOO_SMALL)) {
      free((void*)wvalue);
      wvalue=(CciChar * )malloc(length*sizeof(CciChar)+1);
      memset((void *)wvalue,0x00,(length+1)*sizeof(CciChar));
      length= cniElementName(&ctx->retcode,targetElement,wvalue,length);
   }
   value=(char *)mbString(wvalue,BIP_DEF_COMP_CCSID);
   free((void*)wvalue);
   return value;
}

/*Return the value of a target element */
const CciElementValue  *ElementValueValue(
                                          WmqiContext *ctx,
                                          CciElement *targetElement
                                          ){
   return cniElementValueValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
int   ElementBooleanValue(
                          WmqiContext *ctx,
                          CciElement *targetElement
                          ){
   return cniElementBooleanValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
CciInt   ElementIntegerValue(
                             WmqiContext *ctx,
                             CciElement *targetElement
                             ){
   return cniElementIntegerValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
double   ElementRealValue(
                          WmqiContext *ctx,
                          CciElement *targetElement
                          ){
   return cniElementRealValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
char *   ElementCharacterValue(
                               WmqiContext *ctx,
                               CciElement *targetElement
                               ){
   int length=INIT_CHAR_LENGTH;
   const CciChar * wvalue;
   char *value;
   wvalue=(CciChar * )malloc((length+1)*sizeof(CciChar));
   memset((void*)wvalue,0x00,(length+1)*sizeof(CciChar));
   length=cniElementCharacterValue(&ctx->retcode,targetElement,wvalue,length);
   if ((length>INIT_CHAR_LENGTH)&&(ctx->retcode==CCI_BUFFER_TOO_SMALL)) {
      free((void*)wvalue);
      wvalue=(CciChar * )malloc(length*sizeof(CciChar)+1);
      memset((void *)wvalue,0x00,(length+1)*sizeof(CciChar));
      length= cniElementCharacterValue(&ctx->retcode,targetElement,wvalue,length);
   }
   value=(char *)mbString(wvalue,BIP_DEF_COMP_CCSID);
   free((void*)wvalue);
   return value;
}

/*Return the value of a target element */
struct CciDate   ElementDateValue(
                                  WmqiContext *ctx,
                                  CciElement *targetElement
                                  ){
   return cniElementDateValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
struct CciTime   ElementTimeValue(
                                  WmqiContext *ctx,
                                  CciElement *targetElement
                                  ){
   return cniElementTimeValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
int   ElementByteArrayValue(
                            WmqiContext *ctx,
                            CciElement *targetElement,
                            const struct CciByteArray*value
                            ){
   return cniElementByteArrayValue(&ctx->retcode,targetElement,value);
}

/*Return the value of a target element */
struct CciTimestamp   ElementTimestampValue(
                                            WmqiContext *ctx,
                                            CciElement *targetElement
                                            ){
   return cniElementTimestampValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
struct CciTimestamp   ElementGmtTimestampValue(
                                               WmqiContext *ctx,
                                               CciElement *targetElement
                                               ){
   return cniElementGmtTimestampValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
struct CciTime   ElementGmtTimeValue(
                                     WmqiContext *ctx,
                                     CciElement *targetElement
                                     ){
   return cniElementGmtTimeValue(&ctx->retcode,targetElement);
}

/*Return the value of a target element */
char *   ElementDecimalValue(
                             WmqiContext *ctx,
                             CciElement *targetElement
                             ){
   int length=INIT_CHAR_LENGTH;
   const CciChar * wvalue;
   char *value;
   wvalue=(CciChar * )malloc((length+1)*sizeof(CciChar));
   memset((void *)wvalue,0x00,(length+1)*sizeof(CciChar));
   length=cniElementDecimalValue(&ctx->retcode,targetElement,wvalue,length);
   if ((length>INIT_CHAR_LENGTH)&&(ctx->retcode==CCI_BUFFER_TOO_SMALL)) {
      free((void*)wvalue);
      wvalue=(CciChar * )malloc(length*sizeof(CciChar)+1);
      memset((void *)wvalue,0x00,(length+1)*sizeof(CciChar));
      length= cniElementDecimalValue(&ctx->retcode,targetElement,wvalue,length);
   }
   value=(char *)mbString(wvalue,BIP_DEF_COMP_CCSID);
   free((void*)wvalue);
   return value;
}

/*Return the value of a target element */
int   ElementBitArrayValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           const struct CciBitArray*value
                           ){
   return cniElementBitArrayValue(&ctx->retcode,targetElement,value);
}

/*Copy part of an element tree to a target element */
void   CopyElementTree(
                       WmqiContext *ctx,
                       CciElement *sourceElement,
                       CciElement *targetElement
                       ){
   cniCopyElementTree(&ctx->retcode,sourceElement,targetElement);
}

/*Search the previous sibling chain for a matching element */
CciElement  *SearchPreviousSibling(
                                   WmqiContext *ctx,
                                   CciElement *targetElement,
                                   int mode,
                                   int type,
                                   char   *name
                                   ){
   CciElement *elt;
   const CciChar * wname;
   wname=CciString(name, BIP_DEF_COMP_CCSID) ;
   elt= cniSearchPreviousSibling(&ctx->retcode,targetElement,mode,type,(CciChar *)wname);
   free((void *)wname);
   return elt;
}

/*Search the next sibling chain for a matching element */
CciElement  *SearchNextSibling(
                               WmqiContext *ctx,
                               CciElement *targetElement,
                               int mode,
                               int type,
                               char   *name
                               ){
   CciElement *elt;
   const CciChar * wname;
   wname=CciString(name, BIP_DEF_COMP_CCSID) ;
   elt= cniSearchNextSibling(&ctx->retcode,targetElement,mode,type,(CciChar *)wname);
   free((void *)wname);
   return elt;
}

/*Search the sibling chain from the first child for a matching element */
CciElement  *SearchFirstChild(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              int mode,
                              int type,
                              char   *name
                              ){
   CciElement *elt;
   const CciChar * wname;
   wname=CciString(name, BIP_DEF_COMP_CCSID) ;
   elt= cniSearchFirstChild(&ctx->retcode,targetElement,mode,type,(CciChar *)wname);
   free((void *)wname);
   return elt;
}

/*Search the sibling chain from the last child for a matching element */
CciElement  *SearchLastChild(
                             WmqiContext *ctx,
                             CciElement *targetElement,
                             int mode,
                             int type,
                             char   *name
                             ){
   CciElement *elt;
   const CciChar * wname;
   wname=CciString(name, BIP_DEF_COMP_CCSID) ;
   elt= cniSearchLastChild(&ctx->retcode,targetElement,mode,type,(CciChar *)wname);
   free((void *)wname);
   return elt;
}

/*Return the type attribute of the value of a target element */
int   ElementValueType(
                       WmqiContext *ctx,
                       CciElement *targetElement
                       ){
   return cniElementValueType(&ctx->retcode,targetElement);
}

/*Return the state attribute of the value of a target element */
int   ElementValueState(
                        WmqiContext *ctx,
                        CciElement *targetElement
                        ){
   return cniElementValueState(&ctx->retcode,targetElement);
}

/*Create an SQL statement */
CciSqlExpression  *SqlCreateStatement(
                                      WmqiContext *ctx,
                                      char   *dataSourceName,
                                      int transaction,
                                      char   *statement
                                      ){
   CciSqlExpression *exp;
   const CciChar * wdataSourceName;
   const CciChar * wstatement;
   NODE_CONTEXT_ST * nodectx;
   int rc;
   CCI_EXCEPTION_ST exception_st = {CCI_EXCEPTION_ST_DEFAULT};    
   nodectx=(NODE_CONTEXT_ST *)ctx->nodeContext;
   if (dataSourceName!=0) {
      wdataSourceName=CciString(dataSourceName, BIP_DEF_COMP_CCSID) ;
   } else {
      wdataSourceName=0;
   }
   if (statement!=0) {
      wstatement=CciString(statement, BIP_DEF_COMP_CCSID) ;
   } else {
      wstatement=0;
   }
   
   exp=cniSqlCreateStatement(&ctx->retcode,nodectx->nodeObject,(CciChar *)wdataSourceName,transaction,(CciChar *)wstatement);
   
   
   if (ctx->retcode == CCI_EXCEPTION) {
      /* Get details of the exception */
      memset(&exception_st, 0, sizeof(exception_st));
      cciGetLastExceptionData(&rc, &exception_st);
      printf("Exception %s in statement %ls\n",exception_st.traceText,wstatement);
   }
   free((void *)wdataSourceName);
   
   free((void *)wstatement);

   return exp;
}

/*Execute an SQL expression */
void   SqlExecute(
                  WmqiContext *ctx,
                  CciSqlExpression   *sqlExpression
                  ){
   cniSqlExecute(&ctx->retcode,sqlExpression,ctx->destinationList,ctx->exceptionList,ctx->inputMsg);
}

/*Execute an SQL expression */
void   SqlSelect(
                 WmqiContext *ctx,
                 CciSqlExpression   *sqlExpression,
                 CciMessage   *outputMessage
                 ){
   cniSqlSelect(&ctx->retcode,sqlExpression,ctx->destinationList,ctx->exceptionList,ctx->inputMsg,outputMessage);


}

/*Delete an SQL statement */
void   SqlDeleteStatement(
                          WmqiContext *ctx,
                          CciSqlExpression   *sqlExpression
                          ){
   cniSqlDeleteStatement(&ctx->retcode,sqlExpression);
}

/*Get the environment for a message */
CciMessage  *GetEnvironmentMessage(
                                   WmqiContext *ctx,
                                   CciMessage   *message
                                   ){
   return cniGetEnvironmentMessage(&ctx->retcode,message);
}

/* Propagate a message to a named terminal on the node instance */
int Propagate(
              WmqiContext *ctx,
              CciMessage*  message){
   int retvalue,rc;
   CCI_EXCEPTION_ST exception_st = {CCI_EXCEPTION_ST_DEFAULT};
   if (ctx->outputTerminalObject) {
      if (cniIsTerminalAttached(&rc, ctx->outputTerminalObject)) {
         /* As this is a new, and changed message, it should be finalized... */
         cniFinalize(&rc, message, CCI_FINALIZE_NONE);
         
         retvalue = cniPropagate(&ctx->retcode,ctx->outputTerminalObject,ctx->destinationList,ctx->exceptionList,message);
         if (retvalue == CCI_FAILURE) {
            if (rc == CCI_EXCEPTION) {
               /* Get details of the exception */
               memset(&exception_st, 0, sizeof(exception_st));
               cciGetLastExceptionData(&rc, &exception_st);
               
               /* Any local error handling may go here */
               
               /* Ensure message is deleted prior to return/throw */
               cniDeleteMessage(0, message);
               
               /* We must "rethrow" the exception; note this does not return */
               cciRethrowLastException(&rc);
            }
            else {
               
               /* Some other error...the plugin might choose to log it using the CciLog() */
               /* utility function                                                        */
               
            }
         }
         else {
         }
      }
   }
   else {
      /* Terminal did not exist...severe internal error. The plugin may wish to */
      /* log an error here using the cciLog() utility function.                 */
   }
   return retvalue;
}


/* Create a message object */
CciMessage * CreateMessage(
                           WmqiContext *ctx
                           ) {
   int rc;
   return cniCreateMessage(&ctx->retcode,cniGetMessageContext(&rc, ctx->inputMsg));
}


/* Delete a message object */
void DeleteMessage(
                   WmqiContext *ctx,
                   CciMessage* message
                   ) {
   cniDeleteMessage(&ctx->retcode,message);
   return;
}