/* Perl Wmqi extensions defintions */

#ifndef WMQIPERL
#define WMQIPERL 1
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <bipcni.h>

#define INIT_CHAR_LENGTH 256

typedef struct _WmqiContext {
   CciContext* nodeContext;
   CciMessage* destinationList;
   CciMessage* exceptionList;
   CciMessage* inputMsg;
   CciTerminal *outputTerminalObject;
   int         retcode;
} WmqiContext;

/*Get the root element for a message */
CciElement  *RootElement(
                         WmqiContext *ctx,
                         CciMessage   *message
                         );


/*Create an element before the specified element */
CciElement  *CreateElementBefore(
                                 WmqiContext *ctx,
                                 CciElement *targetElement
                                 );

/*Create an element before the specified element */
CciElement  *CreateElementBeforeUsingParser(
                                            WmqiContext *ctx,
                                            CciElement *targetElement,
                                            char   *parserClassName
                                            );

/*Create an element after the specified element */
CciElement  *CreateElementAfter(
                                WmqiContext *ctx,
                                CciElement *targetElement
                                );

/*Create an element after the specified element */
CciElement  *CreateElementAfterUsingParser(
                                           WmqiContext *ctx,
                                           CciElement *targetElement,
                                           char   *parserClassName
                                           );

/*Create an element as a first child of the specified element */
CciElement  *CreateElementAsFirstChild(
                                       WmqiContext *ctx,
                                       CciElement *targetElement
                                       );

/*Create an element as a first child of the specified element */
CciElement  *CreateElementAsFirstChildUsingParser(
                                                  WmqiContext *ctx,
                                                  CciElement *targetElement,
                                                  char   *parserClassName
                                                  );

/*Create an element as a last child of the specified element */
CciElement  *CreateElementAsLastChild(
                                      WmqiContext *ctx,
                                      CciElement *targetElement
                                      );

/*Create an element as a last child of the specified element */
CciElement  *CreateElementAsLastChildUsingParser(
                                                 WmqiContext *ctx,
                                                 CciElement *targetElement,
                                                 char   *parserClassName
                                                 );

/*Add an element before a target element */
void   AddBefore(
                 WmqiContext *ctx,
                 CciElement *targetElement,
                 CciElement *newElement
                 );

/*Add an element after a target element */
void   AddAfter(
                WmqiContext *ctx,
                CciElement *targetElement,
                CciElement *newElement
                );

/*Add an element as a first child of a target element */
void   AddAsFirstChild(
                       WmqiContext *ctx,
                       CciElement *targetElement,
                       CciElement *newElement
                       );

/*Add an element as a last child of a target element */
void   AddAsLastChild(
                      WmqiContext *ctx,
                      CciElement *targetElement,
                      CciElement *newElement
                      );

/*Detach the target element */
void   Detach(
              WmqiContext *ctx,
              CciElement *targetElement
              );

/*Return the parent of a target element */
CciElement  *Parent(
                    WmqiContext *ctx,
                    CciElement *targetElement
                    );

/*Return the previous sibling of a target element */
CciElement  *PreviousSibling(
                             WmqiContext *ctx,
                             CciElement *targetElement
                             );

/*Return the next sibling of a target element */
CciElement  *NextSibling(
                         WmqiContext *ctx,
                         CciElement *targetElement
                         );

/*Return the first child of a target element */
CciElement  *FirstChild(
                        WmqiContext *ctx,
                        CciElement *targetElement
                        );

/*Return the last child of a target element */
CciElement  *LastChild(
                       WmqiContext *ctx,
                       CciElement *targetElement
                       );

/*Set the type attribute of a target element */
void   SetElementType(
                      WmqiContext *ctx,
                      CciElement *targetElement,
                      int type
                      );

/*Set the name attribute of a target element */
void   SetElementName(
                      WmqiContext *ctx,
                      CciElement *targetElement,
                      char   *name
                      );

/*Set the value attribute of a target element */
void   SetElementValueValue(
                            WmqiContext *ctx,
                            CciElement *targetElement,
                            CciElementValue*value
                            );

/*Set the value of a target element */
void   SetElementBooleanValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              int     value
                              );

/*Set the value of a target element */
void   SetElementIntegerValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              CciInt      value
                              );

/*Set the value of a target element */
void   SetElementRealValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           double     value);

/*Set the value of a target element */
void   SetElementCharacterValue(
                                WmqiContext *ctx,
                                CciElement *targetElement,
                                char   *value,
                                int        length
                                );

/*Set the value of a target element */
void   SetElementDateValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           int year, int month, int day
                           );

/*Set the value of a target element */
void   SetElementTimeValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           int hour, int minute, float second
                           );

/*Set the value of a target element */
void   SetElementByteArrayValue(
                                WmqiContext *ctx,
                                CciElement *targetElement,
                                const struct CciByteArray*value
                                );

/*Set the value of a target element */
void   SetElementTimestampValue(
                                WmqiContext *ctx,
                                CciElement *targetElement,
  int year, int month, int day, int hour, int minute, float second
                                );

/*Set the value of a target element */
void   SetElementGmtTimestampValue(
                                   WmqiContext *ctx,
                                   CciElement *targetElement,
  int year, int month, int day, int hour, int minute, float second
                                   );

/*Set the value of a target element */
void   SetElementGmtTimeValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              int hour, int minute, float second
                              );

/*Set the value of a target element */
void   SetElementDecimalValue(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              char   *value
                              );

/*Set the value of a target element */
void   SetElementBitArrayValue(
                               WmqiContext *ctx,
                               CciElement *targetElement,
                               const struct CciBitArray*value
                               );

/*Return the type attribute of a target element */
int   ElementType(
                  WmqiContext *ctx,
                  CciElement *targetElement
                  );

/*Return the name attribute of a target element */
char *  ElementName(
                  WmqiContext *ctx,
                  CciElement *targetElement
                  );

/*Return the value of a target element */
const CciElementValue  *ElementValueValue(
                                          WmqiContext *ctx,
                                          CciElement *targetElement
                                          );

/*Return the value of a target element */
int   ElementBooleanValue(
                          WmqiContext *ctx,
                          CciElement *targetElement
                          );

/*Return the value of a target element */
CciInt   ElementIntegerValue(
                             WmqiContext *ctx,
                             CciElement *targetElement
                             );

/*Return the value of a target element */
double   ElementRealValue(
                          WmqiContext *ctx,
                          CciElement *targetElement
                          );

/*Return the value of a target element */
char *   ElementCharacterValue(
                            WmqiContext *ctx,
                            CciElement *targetElement
                            );

/*Return the value of a target element */
struct CciDate   ElementDateValue(
                                  WmqiContext *ctx,
                                  CciElement *targetElement
                                  );

/*Return the value of a target element */
struct CciTime   ElementTimeValue(
                                  WmqiContext *ctx,
                                  CciElement *targetElement
                                  );

/*Return the value of a target element */
int   ElementByteArrayValue(
                            WmqiContext *ctx,
                            CciElement *targetElement,
                            const struct CciByteArray*value
                            );

/*Return the value of a target element */
struct CciTimestamp   ElementTimestampValue(
                                            WmqiContext *ctx,
                                            CciElement *targetElement
                                            );

/*Return the value of a target element */
struct CciTimestamp   ElementGmtTimestampValue(
                                               WmqiContext *ctx,
                                               CciElement *targetElement
                                               );

/*Return the value of a target element */
struct CciTime   ElementGmtTimeValue(
                                     WmqiContext *ctx,
                                     CciElement *targetElement
                                     );

/*Return the value of a target element */
char *   ElementDecimalValue(
                          WmqiContext *ctx,
                          CciElement *targetElement
                          );

/*Return the value of a target element */
int   ElementBitArrayValue(
                           WmqiContext *ctx,
                           CciElement *targetElement,
                           const struct CciBitArray*value
                           );

/*Copy part of an element tree to a target element */
void   CopyElementTree(
                       WmqiContext *ctx,
                       CciElement *sourceElement,
                       CciElement *targetElement
                       );

/*Search the previous sibling chain for a matching element */
CciElement  *SearchPreviousSibling(
                                   WmqiContext *ctx,
                                   CciElement *targetElement,
                                   int mode,
                                   int type,
                                   char   *name
                                   );

/*Search the next sibling chain for a matching element */
CciElement  *SearchNextSibling(
                               WmqiContext *ctx,
                               CciElement *targetElement,
                               int mode,
                               int type,
                               char   *name
                               );

/*Search the sibling chain from the first child for a matching element */
CciElement  *SearchFirstChild(
                              WmqiContext *ctx,
                              CciElement *targetElement,
                              int mode,
                              int type,
                              char   *name
                              );

/*Search the sibling chain from the last child for a matching element */
CciElement  *SearchLastChild(
                             WmqiContext *ctx,
                             CciElement *targetElement,
                             int mode,
                             int type,
                             char   *name
                             );

/*Return the type attribute of the value of a target element */
int   ElementValueType(
                       WmqiContext *ctx,
                       CciElement *targetElement
                       );

/*Return the state attribute of the value of a target element */
int   ElementValueState(
                        WmqiContext *ctx,
                        CciElement *targetElement
                        );

/*Create an SQL statement */
CciSqlExpression  *SqlCreateStatement(
                                      WmqiContext *ctx,
                                      char   *dataSourceName,
                                      int transaction,
                                      char   *statement
                                      );

/*Execute an SQL expression */
void   SqlExecute(
                  WmqiContext *ctx,
                  CciSqlExpression   *sqlExpression
                  );

/*Execute an SQL expression */
void   SqlSelect(
                 WmqiContext *ctx,
                 CciSqlExpression   *sqlExpression,
                 CciMessage   *outputMessage
                 );

/*Delete an SQL statement */
void   SqlDeleteStatement(
                          WmqiContext *ctx,
                          CciSqlExpression   *sqlExpression
                          );

/*Get the environment for a message */
CciMessage  *GetEnvironmentMessage(
                                   WmqiContext *ctx,
                                   CciMessage   *message
                                   );


/* Propagate a message to a named terminal on the node instance */
int Propagate(
              WmqiContext *ctx,
              CciMessage*  message
              );

/* Create a message object */
CciMessage * CreateMessage(
                           WmqiContext *ctx
                           );

/* Delete a message object */
void DeleteMessage(
  WmqiContext *ctx,
  CciMessage* message
);	
#endif